icon_gerbview ICON "../bitmaps_png/icons/icon_gerbview.ico"

#define RC_VER_FILE_DESCRIPTION "KiCad Gerber Viewer " KICAD_WIN32_RC_PRODVER_STR
#define RC_VER_INTERNALNAME "gerbview"
#define RC_VER_ORIGINALFILENAME "gerbview.exe"
#include "kiwin32.rc"
